/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <pkgmgr-info.h>
#include <pkgmgr_installer_info.h>
#include <aul.h>

#include "log.h"
#include "utils.h"
#include "pkgmgr_parser_plugin_interface.h"

#include <wait.h>
#include <dirent.h>
#include <sys/stat.h>

#include <algorithm>
#include <string>

#include <pwd.h>
#include <grp.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>

#include <fstream>
#include <sys/smack.h>

#include "ni_common.h"
#include "path_manager.h"
#include "plugin_manager.h"

#ifdef  LOG_TAG
#undef  LOG_TAG
#endif
#define LOG_TAG "NETCORE_INSTALLER_PLUGIN"

#ifndef CROSSGEN_PATH
#error "CROSSGEN_PATH is missed"
#endif

#define __XSTR(x) #x
#define __STR(x) __XSTR(x)
static const char* __CROSSGEN_PATH = __STR(CROSSGEN_PATH);
#undef __STR
#undef __XSTR

static int __interval = 0;
static std::string __tpa;

static void waitInterval()
{
	// by the recommand, ignore small value for performance.
	if (__interval > 10000) {
		fprintf(stderr, "sleep %d usec\n", __interval);
		usleep(__interval);
	}
}

static bool niExist(const std::string& path, std::string& ni)
{
	size_t index = path.find_last_of(".");
	if (index == std::string::npos) {
		return false;
	}
	std::string fName = path.substr(0, index);
	std::string fExt = path.substr(index, path.length());

	// crossgen generate file with lower case extension only
	std::transform(fExt.begin(), fExt.end(), fExt.begin(), ::tolower);
	std::string f = fName + ".ni" + fExt;
	if (isFileExist(f)) {
		ni = f;
		return true;
	}

	// native image of System.Private.CoreLib.dll should have to overwrite
	// original file to support new coreclr
	if (path.find("System.Private.CoreLib.dll") != std::string::npos) {
		std::string coreLibBackup = path + ".Backup";
		if (isFileExist(coreLibBackup)) {
			ni = path;
			return true;
		}
	}

	return false;
}

static void updateNiFileInfo(const std::string& path)
{
	char* label = NULL;
	std::string niPath;

	if (niExist(path, niPath)) {
		// change smack label
		if (smack_getlabel(path.c_str(), &label, SMACK_LABEL_ACCESS) == 0) {
			if (smack_setlabel(niPath.c_str(), label, SMACK_LABEL_ACCESS) < 0) {
				fprintf(stderr, "Fail to set smack label\n");
			}
			free(label);
		}

		// change owner and groups for generated ni file.
		struct stat info;
		if (!stat(path.c_str(), &info)) {
			if (chown(niPath.c_str(), info.st_uid, info.st_gid) == -1)
				fprintf(stderr, "Failed to change owner and group name\n");
		}
	}
}

static void crossgen(const std::string& dllPath, const std::string& appPath, bool enableR2R)
{
	std::string absDllPath = absolutePath(dllPath);

	pid_t pid = fork();
	if (pid == -1)
		return;

	if (pid > 0) {
		int status;
		waitpid(pid, &status, 0);
		if (WIFEXITED(status)) {
			updateNiFileInfo(absDllPath);
			return;
		}
	} else {
		std::string jitPath = getRuntimeDir() + "/libclrjit.so";
		std::vector<const char*> argv = {
			__CROSSGEN_PATH,
			"/nologo",
			"/Trusted_Platform_Assemblies", __tpa.c_str(),
			"/JITPath", jitPath.c_str()
		};

		if (!enableR2R) {
			argv.push_back("/FragileNonVersionable");
		}

		argv.push_back("/App_Paths");
		std::string absAppPath;
		if (!appPath.empty()) {
			absAppPath = appPath;
		} else {
			absAppPath = baseName(absDllPath);
		}
		argv.push_back(absAppPath.c_str());

		argv.push_back(absDllPath.c_str());
		argv.push_back(nullptr);

		fprintf(stderr, "+ %s (%s)\n", absDllPath.c_str(), enableR2R ? "R2R" : "FNV");

		execv(__CROSSGEN_PATH, const_cast<char* const*>(argv.data()));
		exit(0);
	}
}

static int getRootPath(std::string pkgId, std::string& rootPath)
{
	int ret = 0;
	char *path = 0;

	uid_t uid = 0;

	if (pkgmgr_installer_info_get_target_uid(&uid) < 0) {
		_ERR("Failed to get UID");
		return -1;
	}

	pkgmgrinfo_pkginfo_h handle;
	if (uid == 0) {
		ret = pkgmgrinfo_pkginfo_get_pkginfo(pkgId.c_str(), &handle);
		if (ret != PMINFO_R_OK)
			return -1;
	} else {
		ret = pkgmgrinfo_pkginfo_get_usr_pkginfo(pkgId.c_str(), uid, &handle);
		if (ret != PMINFO_R_OK)
			return -1;
	}

	ret = pkgmgrinfo_pkginfo_get_root_path(handle, &path);
	if (ret != PMINFO_R_OK) {
		pkgmgrinfo_pkginfo_destroy_pkginfo(handle);
		return -1;
	}
	rootPath = path;
	pkgmgrinfo_pkginfo_destroy_pkginfo(handle);

	return 0;
}


static int appAotCb(pkgmgrinfo_appinfo_h handle, void *userData)
{
	char *pkgId = NULL;
	int ret = 0;

	ret = pkgmgrinfo_appinfo_get_pkgid(handle, &pkgId);
	if (ret != PMINFO_R_OK) {
		fprintf(stderr, "Failed to get pkgid\n");
		return -1;
	}

	if (removeNiUnderPkgRoot(pkgId) != 0) {
		fprintf(stderr, "Failed to remove previous dlls from [%s]\n", pkgId);
		return -1;
	}

	// Regenerate ni files with R2R mode forcibiliy. (there is no way to now which option is used)
	if (createNiUnderPkgRoot(pkgId, true) != 0) {
		fprintf(stderr, "Failed to get root path from [%s]\n", pkgId);
		return -1;
	} else {
		fprintf(stderr, "Complete make application to native image\n");
	}

	return 0;
}

static void createCoreLibNI(bool enableR2R)
{
	std::string coreLib = concatPath(getRuntimeDir(), "System.Private.CoreLib.dll");
	std::string niCoreLib = concatPath(getRuntimeDir(), "System.Private.CoreLib.ni.dll");
	std::string coreLibBackup = concatPath(getRuntimeDir(), "System.Private.CoreLib.dll.Backup");

	if (!niExist(coreLib, niCoreLib)) {
		crossgen(coreLib, std::string(), enableR2R);
		if (isFileExist(niCoreLib)) {
			if (rename(coreLib.c_str(), coreLibBackup.c_str())) {
				fprintf(stderr, "Failed to rename System.Private.CoreLib.dll\n");
			}
			if (rename(niCoreLib.c_str(), coreLib.c_str())) {
				fprintf(stderr, "Failed to rename System.Private.CoreLib.ni.dll\n");
			}
		} else {
			fprintf(stderr, "Failed to create native image for %s\n", coreLib.c_str());
		}
	}
}

int initNICommon(NiCommonOption* option)
{
#if defined(__arm__)
	// get interval value
	const char* intervalFile = "/usr/share/dotnet.tizen/lib/crossgen_interval.txt";
	std::ifstream inFile(intervalFile);
	if (inFile) {
		fprintf(stderr, "crossgen_interval.txt is found\n");
		inFile >> __interval;
	}

	if (initializePluginManager("normal")) {
		fprintf(stderr, "Fail to initialize plugin manager\n");
		return -1;
	}
	if (initializePathManager(option->runtimeDir, option->tizenFXDir, option->extraDirs)) {
		fprintf(stderr, "Fail to initialize path manager\n");
		return -1;
	}

	__tpa = getTPA();

	return 0;
#else
	fprintf(stderr, "crossgen supports arm architecture only. skip ni file generation\n");
	return -1;
#endif
}

void finalizeNICommon()
{
	__interval = 0;

	finalizePluginManager();
	finalizePathManager();

	__tpa.clear();
}


void createNiPlatform(bool enableR2R)
{
	const std::string platformDirs[] = {getRuntimeDir(), getTizenFXDir()};

	createNiUnderDirs(platformDirs, 2, enableR2R);
}

void createNiDll(const std::string& dllPath, bool enableR2R)
{
	createCoreLibNI(enableR2R);

	if (!isFileExist(dllPath)) {
		fprintf(stderr, "Failed to find dll : %s\n", dllPath.c_str());
		return;
	}

	std::string niPath;
	if (niExist(dllPath, niPath)) {
		fprintf(stderr, "Already [%s] file is exist\n", niPath.c_str());
		return;
	}

	crossgen(dllPath, std::string(), enableR2R);
	if (!niExist(dllPath, niPath)) {
		fprintf(stderr, "Failed to create native image for %s\n", dllPath.c_str());
	}
}

void createNiUnderDirs(const std::string rootPaths[], int count, const std::string ignores[], int igcount, afterCreate cb, bool enableR2R)
{
	createCoreLibNI(enableR2R);

	std::string appPaths;
	for (int i = 0; i < count; i++) {
		appPaths += rootPaths[i];
		appPaths += ':';
	}

	if (appPaths.back() == ':')
		appPaths.pop_back();

	auto convert = [&appPaths, ignores, igcount, &cb, enableR2R](const std::string& path, const char* name) {
		for (int i = 0; i < igcount; i++) {
			if (path == ignores[i])
				return;
		}
		std::string niPath;
		if (isManagedAssembly(path)) {
			if (niExist(path, niPath)) {
				fprintf(stderr, "Already [%s] file is exist\n", niPath.c_str());
				return;
			}
			crossgen(path, appPaths.c_str(), enableR2R);
			if (niExist(path, niPath)) {
				if (cb != nullptr)
					cb(niPath.c_str());
			} else {
				fprintf(stderr, "Failed to create native image for %s\n", path.c_str());
			}
			waitInterval();
		}
	};

	for (int i = 0; i < count; i++)
		scanFilesInDir(rootPaths[i], convert, 1);
}
void createNiUnderDirs(const std::string rootPaths[], int count, afterCreate cb, bool enableR2R)
{
	createNiUnderDirs(rootPaths, count, nullptr, 0, cb, enableR2R);
}
void createNiUnderDirs(const std::string rootPaths[], int count, bool enableR2R)
{
	createNiUnderDirs(rootPaths, count, nullptr, enableR2R);
}

int createNiUnderPkgRoot(const std::string& pkgName, bool enableR2R)
{
	std::string pkgRoot;
	if (getRootPath(pkgName, pkgRoot) < 0)
		return -1;

	std::string binDir = concatPath(pkgRoot, "bin");
	std::string libDir = concatPath(pkgRoot, "lib");
	std::string paths[] = {binDir, libDir};

	createNiUnderDirs(paths, 2, enableR2R);

	return 0;
}

int createNiDllUnderPkgRoot(const std::string& pkgName, const std::string& dllPath, bool enableR2R)
{
	std::string pkgRoot;
	if (getRootPath(pkgName, pkgRoot) < 0)
		return -1;

	std::string binDir = concatPath(pkgRoot, "bin");
	std::string libDir = concatPath(pkgRoot, "lib");
	std::string appPaths = binDir + ":" + libDir;

	if (!isFileExist(dllPath)) {
		fprintf(stderr, "Failed to find dll : %s\n", dllPath.c_str());
		return -1;
	}

	std::string niPath;
	if (niExist(dllPath, niPath)) {
		fprintf(stderr, "Already [%s] file is exist\n", niPath.c_str());
		return -1;
	}

	crossgen(dllPath, appPaths, enableR2R);
	if (!niExist(dllPath, niPath)) {
		fprintf(stderr, "Failed to create native image for %s\n", dllPath.c_str());
		return -1;
	}

	return 0;
}

void removeNiPlatform()
{
	std::string coreLib = concatPath(getRuntimeDir(), "System.Private.CoreLib.dll");
	std::string coreLibBackup = concatPath(getRuntimeDir(), "System.Private.CoreLib.dll.Backup");

	if (!isFileExist(coreLibBackup)) {
		return;
	}

	if (remove(coreLib.c_str())) {
		fprintf(stderr, "Failed to remove System.Private.CoreLib native image file\n");
	}

	if (rename(coreLibBackup.c_str(), coreLib.c_str())) {
		fprintf(stderr, "Failed to rename System.Private.CoreLib.Backup to origin\n");
	}

	const std::string platformDirs[] = {getRuntimeDir(), getTizenFXDir()};

	removeNiUnderDirs(platformDirs, 2);
}

void removeNiUnderDirs(const std::string rootPaths[], int count)
{
	auto convert = [](const std::string& path, std::string name) {
		std::string ni;
		if (isNativeImage(path)) {
			if (remove(path.c_str())) {
				fprintf(stderr, "Failed to remove %s\n", path.c_str());
			}
		}
	};

	for (int i = 0; i < count; i++)
		scanFilesInDir(rootPaths[i], convert, -1);
}

int removeNiUnderPkgRoot(const std::string& pkgName)
{
	std::string pkgRoot;
	if (getRootPath(pkgName, pkgRoot) < 0)
		return -1;

	std::string binDir = concatPath(pkgRoot, "bin");
	std::string libDir = concatPath(pkgRoot, "lib");
	std::string paths[] = {binDir, libDir};

	removeNiUnderDirs(paths, 2);

	return 0;
}

int regenerateAppNI()
{
	int ret = 0;
	pkgmgrinfo_appinfo_metadata_filter_h handle;

	ret = pkgmgrinfo_appinfo_metadata_filter_create(&handle);
	if (ret != PMINFO_R_OK)
		return -1;

	ret = pkgmgrinfo_appinfo_metadata_filter_add(handle, "http://tizen.org/metadata/prefer_dotnet_aot", "true");
	if (ret != PMINFO_R_OK) {
		pkgmgrinfo_appinfo_metadata_filter_destroy(handle);
		return -1;
	}

	ret = pkgmgrinfo_appinfo_metadata_filter_foreach(handle, appAotCb, NULL);
	if (ret != PMINFO_R_OK) {
		fprintf(stderr, "Failed pkgmgrinfo_appinfo_metadata_filter_foreach\n");
		pkgmgrinfo_appinfo_metadata_filter_destroy(handle);
		return -1;
	}

	fprintf(stderr, "Success pkgmgrinfo_appinfo_metadata_filter_foreach\n");

	pkgmgrinfo_appinfo_metadata_filter_destroy(handle);
	return 0;
}

